// -*- C++ -*-
#define __USE_XOPEN2K8
#include <iostream>
#include <sstream>
#include <string>

#include "Rivet/Analysis.hh"
#include "Rivet/RivetAIDA.hh"
#include "Rivet/Tools/Logging.hh"
#include "Rivet/Projections/FinalState.hh"

#include "Rivet/Projections/LeadingParticlesFinalState.hh"
#include "Rivet/Jet.hh"
#include "Rivet/Projections/FastJets.hh"

#include "fastjet/internal/base.hh"
#include "fastjet/JetDefinition.hh"
#include "fastjet/AreaDefinition.hh"
#include "fastjet/ClusterSequence.hh"
#include "fastjet/ClusterSequenceArea.hh"
#include "fastjet/PseudoJet.hh"


namespace Rivet {


  class ATLAS_2011_I921594 : public Analysis {
  public:

    /// Constructor
    ATLAS_2011_I921594()
      : Analysis("ATLAS_2011_I921594")
    {
      setBeams(PROTON, PROTON);
      setNeedsCrossSection(true);

      _eta_bins.push_back( 0.00);
      _eta_bins.push_back( 0.60);
      _eta_bins.push_back( 1.37);
      _eta_bins.push_back( 1.52);
      _eta_bins.push_back( 1.81);
      _eta_bins.push_back( 2.37);

      _eta_bins_areaoffset.push_back(0.0);
      _eta_bins_areaoffset.push_back(1.5);
      _eta_bins_areaoffset.push_back(3.0);
    }

  public:

    /// Book histograms and initialise projections before the run
    void init() {

      FinalState fs;
      addProjection(fs, "FS");

      // consider the final state jets for the energy density calculation
      FastJets fj(fs, FastJets::KT, 0.5);
      _area_def = new fastjet::AreaDefinition(fastjet::VoronoiAreaSpec());
      fj.useJetArea(_area_def);
      addProjection(fj, "KtJetsD05");

      // consider the leading pt photon with |eta|<2.37 and pT>45 GeV
      LeadingParticlesFinalState photonfs(FinalState(-2.37, 2.37, 45.0*GeV));
      photonfs.addParticleId(PHOTON);
      addProjection(photonfs, "LeadingPhoton");

      // book the dsigma/dEt (in eta bins) histograms
      for(int i=0; i < _eta_bins.size() - 1; i++) {
        if (fabs(_eta_bins[i] - 1.37) < .0001) continue;
        _h_Et_photon[i] = bookHistogram1D(1, 1, i+1);
      }
    }

    /// return eta bin for either dsigma/dET histogram (area_eta=false) or energy density correction (area_eta=true)
    int getEtaBin(double eta_w, bool area_eta) const {

      double eta = fabs(eta_w);

      int v_iter = 0;

      if ( ! area_eta) {
	for (v_iter=0; v_iter < (int)_eta_bins.size() - 1; v_iter++) {
	  if (eta >= _eta_bins.at(v_iter) && eta < _eta_bins.at(v_iter+1))   break;
	}
      }
      else {
	for (v_iter=0; v_iter < (int)_eta_bins_areaoffset.size() - 1; v_iter++) {
	  if (eta >= _eta_bins_areaoffset.at(v_iter) && eta < _eta_bins_areaoffset.at(v_iter+1))   break;
	}
      }
      
      return v_iter;
    }


    /// Perform the per-event analysis
    void analyze(const Event& event) {

      // retrieve event weight
      const double weight = event.weight();

      // retrieve leading photon
      ParticleVector photons = applyProjection<LeadingParticlesFinalState>(event, "LeadingPhoton").particles();

      if (photons.size() != 1) {
	vetoEvent;
      }

      // retrieve 4-momentum of leading photon
      FourMomentum leadingPhoton = photons[0].momentum();
      double eta_P = leadingPhoton.eta();
      double phi_P = leadingPhoton.phi();
      double et_P =  leadingPhoton.Et();
      
      // veto events with photon in crack
      if (fabs(eta_P) >= 1.37 && fabs(eta_P) < 1.52) {
	vetoEvent;
      }

      // compute isolation energy in cone of radius .4 around photon (all particles)
      ParticleVector fs = applyProjection<FinalState>(event, "FS").particles();
      FourMomentum mom_in_EtCone;
      foreach (const Particle& p, fs) {
	// check if it's in the cone of .4
        if (deltaR(eta_P, phi_P, p.momentum().eta(), p.momentum().phi()) >= 0.4) continue;

	// check if it's in the 5x7 central core
	if (fabs(eta_P-p.momentum().eta()) < .025*7.0*0.5 && 
	    fabs(phi_P-p.momentum().phi()) < (PI/128.)*5.0*0.5) continue;

	mom_in_EtCone += p.momentum();
      }

      // compute the median energy density
      _ptDensity.clear();
      _sigma.clear();
      _Njets.clear();

      std::vector< std::vector<double> > ptDensities;
      std::vector<double> emptyVec;
      ptDensities.assign(_eta_bins_areaoffset.size()-1,emptyVec);

      const fastjet::ClusterSequenceArea* clust_seq_area = applyProjection<FastJets>(event, "KtJetsD05").clusterSeqArea();
      foreach (const fastjet::PseudoJet& jet, applyProjection<FastJets>(event, "KtJetsD05").pseudoJets(0.0*GeV)) {
        double y   = fabs(jet.rapidity());
	double eta = fabs(jet.eta());
	double pt  = fabs(jet.perp());

	// get the cluster sequence
	double area = clust_seq_area->area(jet);

	if (area > 10e-4 && fabs(eta) < _eta_bins_areaoffset[_eta_bins_areaoffset.size()-1]) {
	  ptDensities.at( getEtaBin(fabs(eta),true) ).push_back(pt/area);
	}
      }

      for (int b=0; b < (int)_eta_bins_areaoffset.size() - 1; b++) {
	double median = 0.0;
	double sigma  = 0.0;
	int Njets = 0;

	if (ptDensities[b].size() > 0) {
	  std::sort( ptDensities[b].begin(), ptDensities[b].end());
	  int nDens = ptDensities[b].size();
	  if( nDens % 2 == 0 )
	    median = (ptDensities[b][nDens/2] + ptDensities[b][(nDens-2)/2]) / 2;
	  else
	    median = ptDensities[b][(nDens-1)/2];
	  sigma = ptDensities[b][(int)(.15865*nDens)];
	  Njets = nDens;
	}
	_ptDensity.push_back(median);
	_sigma.push_back(sigma);
	_Njets.push_back(Njets);
      }

      // compute the isolation energy correction (cone area*energy density)
      float EtCone_area = PI*.4*.4 - (7.0*.025)*(5.0*PI/128.);
      float correction = _ptDensity[getEtaBin(eta_P,true)]*EtCone_area;

      // apply isolation cut
      if (mom_in_EtCone.Et() - correction > 4.0*GeV) { 
	vetoEvent;
      }

      // fill histograms
      int eta_bin = getEtaBin(eta_P,false);
      _h_Et_photon[eta_bin]->fill(et_P, weight);
    }


    /// Normalise histograms etc., after the run
    void finalize() {

      for (int i=0; i<_eta_bins.size() - 1; i++) {
	if (fabs(_eta_bins[i] - 1.37) < .0001) continue;
	scale(_h_Et_photon[i], crossSection()/sumOfWeights());
      }
    }

  private:

    AIDA::IHistogram1D *_h_Et_photon[5];

    fastjet::AreaDefinition* _area_def;

    std::vector<float> _eta_bins;
    std::vector<float> _eta_bins_areaoffset;

    std::vector<float> _ptDensity;
    std::vector<float> _sigma;
    std::vector<float> _Njets;
  };


  // This global object acts as a hook for the plugin system
  DECLARE_RIVET_PLUGIN(ATLAS_2011_I921594);
}
